from bestia.iterate import looped_list_item
from bestia.output import echo

from kord.errors import *

SHARP = 1
FLAT = -1

TONE = 2
SEMITONE = 1

UNISON = 0
DIMINISHED_SECOND = 0

MINOR_SECOND = 1
AUGMENTED_UNISON = 1

MAJOR_SECOND = 2
DIMINISHED_THIRD = 2

MINOR_THIRD = 3
AUGMENTED_SECOND = 3

DIMINISHED_FOURTH = 4
MAJOR_THIRD = 4

PERFECT_FOURTH = 5
AUGMENTED_THIRD = 5

AUGMENTED_FOURTH = 6
DIMINISHED_FIFTH = 6

PERFECT_FIFTH = 7
DIMINISHED_SIXTH = 7

MINOR_SIXTH = 8
AUGMENTED_FIFTH = 8

MAJOR_SIXTH = 9
DIMINISHED_SEVENTH = 9

MINOR_SEVENTH = 10
AUGMENTED_SIXTH = 10

MAJOR_SEVENTH = 11
DIMINISHED_OCTAVE = 11

OCTAVE = 12
AUGMENTED_SEVENTH = 12

_TONES = (
    'C',
    None,
    'D',
    None,
    'E',
    'F',
    None,
    'G',
    None,
    'A',
    None,
    'B',
)

_OCTS = (
    # '₀',
    # '₁',
    # '₂',
    # '₃',
    # '₄',
    # '₅',
    # '₆',
    # '₇',
    # '₈',
    # '₉',
    '⁰',
    '¹',
    '²',
    '³',
    '⁴',
    '⁵',
    '⁶',
    '⁷',
    '⁸',
    '⁹',
)

_ALTS = {
    'bb': '𝄫',
    'b': '♭',
    # 'b': 'ᵇ',
    '': '',
    # '#': '⌗',
    # '#': '⋕',
    '#': '♯',
    '##': '𝄪',
}

def input_alterations():
    return list(_ALTS.keys())

def output_alterations():
    return list(_ALTS.values())

class Note(object):


    def __init__(self, *args):
        self.tone = args[0].upper()
        if self.tone not in _TONES:
            raise InvalidTone(args[0])

        self.alt = ''
        for a in args[1:]:
            if a in input_alterations():
                self.alt = a
        if self.alt not in input_alterations():
            raise InvalidAlteration(self.alt)

        self.oct = 3
        for a in args[1:]:
            if type(a) == int:
                self.oct = a
            # raise InvalidOctave(self.oct)


    def __iter__(self):
        for i in (self.tone, self.alt, self.oct):
            yield i


    def __repr__(self):
        return '{}{}{}'.format(
            self.tone,
            self.repr_alt,
            self.repr_oct,
        )

    @property
    def repr_oct(self):
        output = ''
        for c in str(self.oct):
            output += _OCTS[int(c)]
        return output

    @property
    def repr_alt(self):
        return _ALTS[self.alt]


    def __eq__(self, other):
        return self.delta_semitones(other) == 0

    def __gt__(self, other):
        return self.delta_semitones(other) > 0

    def __ge__(self, other):
        return self.delta_semitones(other) >= 0

    def is_note(self, other, ignore_oct=False):
        ''' while the arithmetic methods above strictly compare
            delta semitones, this method compares Note attributes:
                * tone
                * alt
                * oct (optional)
        '''
        if self.__class__ == type(other):
            if self.tone == other.tone:
                if self.alt == other.alt:
                    if ignore_oct:
                        return True
                    return self.oct == other.oct


    ### TONE METHODS
    def relative_tone(self, n):
        my_index = _TONES.index(self.tone)
        return looped_list_item(my_index +n, _TONES)

    def adjacent_tone(self, n=1):
        ''' returns next adjacent tone of self
            negative fails...
        '''
        tone = None
        tone_count = 0
        i = 1

        while True:
            if tone_count == n:
                return tone

            tone = self.relative_tone(i)
            if tone:
                tone_count += 1

            i += 1


    def delta_semitones(self, other):
        # 0  :  self == other
        # >1 :  self >  other
        # <1 :  self  < other
        oct_delta = self.oct - other.oct
        oct_delta_st = oct_delta * OCTAVE
        tone_delta_st = _TONES.index(self.tone) - _TONES.index(other.tone)
        alt_delta_st  = input_alterations().index(self.alt) - input_alterations().index(other.alt)
        return oct_delta_st + tone_delta_st + alt_delta_st

    # ENHARMONIC ATTRIBUTES
    @property
    def enharmonic_row(self):
        return self.matrix_coordinates[0]

    @property
    def matrix_coordinates(self):
        for _row_index, _row in enumerate(ENHARMONIC_MATRIX):
            for _note_index, _enharmonic_note in enumerate(_row):
                if self.tone == _enharmonic_note.tone and self.alt == _enharmonic_note.alt:
                    return (_row_index, _note_index)


### This is the heart of the whole project
### indeces are used to determine:
###   * when to change octs
###   * delta_semitones between tones
### notes MUST be unique so that Note.degree() finds 1 exact match!
ENHARMONIC_MATRIX = (

    ## 2-octave enharmonic relationships
    (  Note('C', '' , 2), Note('B', '#' , 1), Note('D', 'bb', 2)  ), # NAH
    (  Note('C', '#', 2), Note('D', 'b' , 2), Note('B', '##', 1)  ), # AAH

    ## 1-octave enharmonic relationships
    (  Note('D', '' , 1), Note('C', '##', 1), Note('E', 'bb', 1)  ), # NHH
    (  Note('D', '#', 1), Note('E', 'b' , 1), Note('F', 'bb', 1)  ), # AAH
    (  Note('E', '' , 1), Note('F', 'b' , 1), Note('D', '##', 1)  ), # NAH
    (  Note('F', '' , 1), Note('E', '#' , 1), Note('G', 'bb', 1)  ), # NAH
    (  Note('F', '#', 1), Note('G', 'b' , 1), Note('E', '##', 1)  ), # AAH
    (  Note('G', '' , 1), Note('F', '##', 1), Note('A', 'bb', 1)  ), # NHH
    (  Note('G', '#', 1), Note('A', 'b' , 1)                      ), # AA
    (  Note('A', '' , 1), Note('G', '##', 1), Note('B', 'bb', 1)  ), # NHH

    ## 2-octave enharmonic relationships
    (  Note('A', '#', 1), Note('B', 'b' , 1), Note('C', 'bb', 2)  ), # AAH
    (  Note('B', '' , 1), Note('C', 'b' , 2), Note('A', '##', 1)  ), # NAH

)



# from kord import universe
# from kord import tunings
# from kord import keys
# from kord import notes
# from kord import errors
from kord.universe import *
from kord.tunings import *
from kord.keys import *
from kord.notes import *
from kord.errors import *



from collections import deque

from bestia.output import Row, FString, echo
from bestia.iterate import looped_list_item

from kord.notes import *

class TonalKey(object):

    def __init__(self, tone, alt='', oct=0):
        # Major F##, C## work - G## dies...
        self.root = Note(tone, alt, 0)


    def __repr__(self):
        spell_line = Row()
        for d in self.scale(
            notes=len(self._intervals) +1, yield_all=False
        ):
            spell_line.append(
                FString(d, size=5)
            )
        return str(spell_line)


    def interval_from_root(self, d):
        ''' return degree's delta semitones from key's root '''
        if d > len(self._intervals):
            return self.interval_from_root(
                d - len(self._intervals)
            ) + OCTAVE
        return self._intervals[d -1]


    def _spell(self, notes=0, start_note=None, yield_all=True, filter_degrees=[]):

        notes_to_yield = notes if notes else len(self._intervals)
        start_note = start_note if start_note else self.root
        filter_degrees = deque(
            [ self.degree(d) for d in filter_degrees ]
        ) if filter_degrees else []

        yield_enabled = False
        d = 0
        while notes_to_yield:

            d += 1 # ignore 0

            degree = self.degree(d)
            if not degree:
                raise InvalidChord()

            # DETERMINE WHETHER THRESHOLD_NOTE HAS BEEN REACHED
            if not yield_enabled and degree >= start_note:
                yield_enabled = True

                # ROTATE FILTER_DEGREES TO APPROPRIATE_NOTE
                for fd in filter_degrees:
                    if Note(fd.tone, fd.alt) >= Note(degree.tone, degree.alt):
                        filter_degrees.rotate(
                            0 - filter_degrees.index(fd)
                        )
                        break

            if not yield_enabled:
                continue

            # CALCULATE AND YIELD NON-DIATONIC SEMITONES
            last_note_delta = self.interval_from_root(d) - self.interval_from_root(d -1)
            if last_note_delta > SEMITONE:
                for st in range(last_note_delta -1):
                    if yield_all and degree != start_note:
                        # AVOID YIELDING EXTRA NONE BEFORE START_NOTE
                        # WHEN SCALE DEG BEFORE IS > 1ST AWAY
                        yield None

            # DETERMINE WHETHER TO YIELD A DEGREE OR NOT
            yield_note = False if filter_degrees else True
            if filter_degrees:
                if degree.is_note(filter_degrees[0], ignore_oct=True):
                    yield_note = True
                    filter_degrees.rotate(-1)

            if yield_note:
                yield degree
                notes_to_yield -= 1
            else:
                if yield_all:
                    yield None

    def scale(self, notes=0, start_note=None, yield_all=True):
        return self._spell(
            notes=notes, start_note=start_note,
            yield_all=yield_all, filter_degrees=range(1, len(self._intervals) +1),
        )

    # def scale(self, notes=0, start_note=None, yield_all=True):
    #     ''' document better........
    #     yields Notes for diatonic degrees
    #     if all is set, Nones are yield for empty semi-tones '''

    #     notes_to_yield = notes if notes else len(self._intervals)
    #     start_note = start_note if start_note else self.root

    #     yield_enabled = False
    #     d = 0
    #     while notes_to_yield:

    #         d += 1 # ignore 0

    #         degree = self.degree(d)
    #         if not degree:
    #             raise InvalidScale(
    #                 '{}{} {}'.format(
    #                     self.root.tone,
    #                     self.root.repr_alt,
    #                     self.__class__.__name__,
    #                 )
    #             )

    #         if not yield_enabled and degree >= start_note:
    #             yield_enabled = True
    #         # DETERMINE WHETHER THRESHOLD_NOTE HAS BEEN REACHED
    #         if not yield_enabled:
    #             continue

    #         last_note_delta = self.interval_from_root(d) - self.interval_from_root(d -1)
    #         if last_note_delta > SEMITONE:
    #             for st in range(last_note_delta -1):
    #                 if yield_all and degree != start_note:
    #                     # AVOID YIELDING EXTRA NONE BEFORE START_NOTE
    #                     # WHEN SCALE DEG BEFORE IS > 1ST AWAY
    #                     yield None

    #         notes_to_yield -= 1
    #         yield degree


class DiatonicKey(TonalKey):

    def degree(self, d):

        if d == 1:
            return self.root

        # GET DEGREE's ROOT OFFSETS = OCTAVE + SPARE_STS
        octs_from_root, spare_sts = divmod(
            self.interval_from_root(d), OCTAVE
        )

        # GET DEGREE PROPERTIES FROM ENHARMONIC MATRIX
        next_degrees = [
            n for n in looped_list_item(
                self.root.enharmonic_row + spare_sts,
                ENHARMONIC_MATRIX
            ) if n.tone == self.root.adjacent_tone(d -1) # EXPECTED TONE
        ]

        if len(next_degrees) == 1:
            deg = next_degrees[0]
            # if deg.tone == 'C': increase_oct()
        
            # RETURN NEW OBJECT, DO NOT CHANGE OCT OF ENHARMONIC_MATRIX ITEM!
            return Note(
                deg.tone,
                deg.alt,
                octs_from_root if deg.enharmonic_row >= self.root.enharmonic_row else octs_from_root +1
            )


    def triad(self, notes=0, start_note=None, yield_all=True):
        return self._spell(
            notes=notes, start_note=start_note,
            yield_all=yield_all, filter_degrees=(1, 3, 5),
        )

    def seventh(self, notes=0, start_note=None, yield_all=True):
        return self._spell(
            notes=notes, start_note=start_note,
            yield_all=yield_all, filter_degrees=(1, 3, 5, 7),
        )

    def ninth(self, notes=0, start_note=None, yield_all=True):
        return self._spell(
            notes=notes, start_note=start_note,
            yield_all=yield_all, filter_degrees=(1, 3, 5, 7, 9),
        )

    def eleventh(self, notes=0, start_note=None, yield_all=True):
        return self._spell(
            notes=notes, start_note=start_note,
            yield_all=yield_all, filter_degrees=(1, 3, 5, 7, 9, 11),
        )

    def thirteenth(self, notes=0, start_note=None, yield_all=True):
        return self._spell(
            notes=notes, start_note=start_note,
            yield_all=yield_all, filter_degrees=(1, 3, 5, 7, 9, 11, 13),
        )


########################
### MAJOR KEYS/MODES ###
########################

class MajorKey(DiatonicKey):

    _intervals = (
        UNISON,
        MAJOR_SECOND,
        MAJOR_THIRD,
        PERFECT_FOURTH,
        PERFECT_FIFTH,
        MAJOR_SIXTH,
        MAJOR_SEVENTH,
    )

class IonianMode(MajorKey):
    pass

class MixolydianMode(MajorKey):

    _intervals = (
        UNISON,
        MAJOR_SECOND,
        MAJOR_THIRD,
        PERFECT_FOURTH,
        PERFECT_FIFTH,
        MAJOR_SIXTH,
        MINOR_SEVENTH, # <<<
    )

class LydianMode(MajorKey):

    _intervals = (
        UNISON,
        MAJOR_SECOND,
        MAJOR_THIRD,
        AUGMENTED_FOURTH, # <<<
        PERFECT_FIFTH,
        MAJOR_SIXTH,
        MAJOR_SEVENTH,
    )

    # bla = [
    #     TONE, 
    #     SEMITONE,
    #     TONE, 
    # ]

########################
### MINOR KEYS/MODES ###
########################

class MinorKey(DiatonicKey):

    _intervals = (
        UNISON,
        MAJOR_SECOND,
        MINOR_THIRD,
        PERFECT_FOURTH,
        PERFECT_FIFTH,
        MINOR_SIXTH,
        MINOR_SEVENTH,
    )
    
class NaturalMinorKey(MinorKey):
    pass

class MelodicMinorKey(MinorKey):

    _intervals = (
        UNISON,
        MAJOR_SECOND,
        MINOR_THIRD,
        PERFECT_FOURTH,
        PERFECT_FIFTH,
        MAJOR_SIXTH, # <<<
        MAJOR_SEVENTH, # <<<
    )

class HarmonicMinorKey(MinorKey):

    _intervals = (
        UNISON,
        MAJOR_SECOND,
        MINOR_THIRD,
        PERFECT_FOURTH,
        PERFECT_FIFTH,
        MINOR_SIXTH,
        MAJOR_SEVENTH, # <<<
    )

class AeolianMode(MinorKey):
    pass

class DorianMode(MinorKey):

    _intervals = (
        UNISON,
        MAJOR_SECOND,
        MINOR_THIRD,
        PERFECT_FOURTH,
        PERFECT_FIFTH,
        MAJOR_SIXTH, # <<<
        MINOR_SEVENTH,
    )

class PhrygianMode(MinorKey):

    _intervals = (
        UNISON,
        MINOR_SECOND, # <<<
        MINOR_THIRD,
        PERFECT_FOURTH,
        PERFECT_FIFTH,
        MINOR_SIXTH,
        MINOR_SEVENTH,
    )



class ChromaticKey(TonalKey):

    _intervals = (
        UNISON,
        MINOR_SECOND,
        MAJOR_SECOND,
        MINOR_THIRD,
        MAJOR_THIRD,
        PERFECT_FOURTH,
        AUGMENTED_FOURTH,
        PERFECT_FIFTH,
        MINOR_SIXTH,
        MAJOR_SIXTH,
        MINOR_SEVENTH,
        MAJOR_SEVENTH,
    )

    def degree(self, d):

        if d == 1:
            return self.root

        # GET DEGREE's ROOT OFFSETS = OCTAVE + SPARE_STS
        octs_from_root, spare_sts = divmod(
            self.interval_from_root(d), OCTAVE
        )

        # GET DEGREE PROPERTIES FROM ENHARMONIC MATRIX
        # DO I REALLY NEED THESE 3 CHECKS ?
        # MATCH ROOT_TONE
        next_degrees = [
            n for n in looped_list_item(
                self.root.enharmonic_row + spare_sts,
                ENHARMONIC_MATRIX
            ) if n.is_note(self.root, ignore_oct=True)
        ]

        if not next_degrees:
            # MATCH ROOT_ALT
            next_degrees = [
                n for n in looped_list_item(
                    self.root.enharmonic_row + spare_sts,
                    ENHARMONIC_MATRIX
                ) if n.alt == self.root.alt[:-1]
            ]

            if not next_degrees:
                # CHOOSE "#" or ""
                chosen_alt = '#' if self.root.alt == '' else self.root.alt
                next_degrees = [
                    n for n in looped_list_item(
                        self.root.enharmonic_row + spare_sts,
                        ENHARMONIC_MATRIX
                    ) if n.alt == chosen_alt
                ]

        if len(next_degrees) == 1:

            deg = next_degrees[0] # got from ENH_MATRIX
            # if Note(deg.tone, deg.alt) == Note('C'): increase_oct()

            # RETURN NEW OBJECT, DO NOT CHANGE OCT OF ENHARMONIC_MATRIX ITEM!
            return Note(
                deg.tone,
                deg.alt,
                octs_from_root if deg.enharmonic_row >= self.root.enharmonic_row else octs_from_root +1
            )



# THESE MODULE APPROPRIATELY TESTS LOWER-LEVEL MODULES BEFORE HITTING RUN-TIME
from bestia.output import echo

from kord.tunings import *

EQUALS = (

    (Note('c'), Note('c')),
    (Note('d'), Note('d')),
    (Note('e'), Note('e')),
    (Note('f'), Note('f')),
    (Note('g'), Note('g')),
    (Note('a'), Note('a')),
    (Note('b'), Note('b')),

    (Note('A', '#'), Note('B', 'b')),
    (Note('A', '##'), Note('B', '')),
    (Note('C', ''), Note('D', 'bb')),
    (Note('C', '#'), Note('D', 'b')),

)


NON_EQUALS = (

    (Note('C', 'b', 3), Note('B', '#', 3)),

    (Note('C', 'b', 3), Note('B', '#', 3)),

    (Note('E', 'b', 5), Note('D', '#', 4)),

    (Note('B', '', 5), Note('C', 'b', 4)),

    (Note('E', '#', 3), Note('F', '', 4)),


    (Note('B', 'b'), Note('C', 'bb')),
    (Note('A', '#'), Note('C', 'bb')),

    (Note('B'     ), Note('C', 'b')),
    (Note('A', '##'), Note('C', 'b')),

    (Note('B', '#'), Note('C', '')),
    (Note('B', '#'), Note('D', 'bb')),

    (Note('B', '##'), Note('C', '#')),
    (Note('B', '##'), Note('D', 'b')),


    # these should eval False OK
    (Note('A', '#'), Note('B', 'b', 4)),
    (Note('A', '##'), Note('B', '', 4)),
    (Note('C', ''), Note('D', 'bb', 4)),
    (Note('C', '#'), Note('D', 'b', 4)),
)


def equality_test(ls):
    for l in ls:
        echo('{} == {}\t{}\t{} == {}\t{}'.format(
            l[0], l[1], l[0] == l[1],
            l[1], l[0], l[1] == l[0]),            
            'green'
        )
        assert l[0] == l[1]
        assert l[1] == l[0]

def inequality_test(ls):
    for l in ls:
        echo('{} != {}\t{}\t{} != {}\t{}'.format(
            l[0], l[1], l[0] != l[1],
            l[1], l[0], l[1] != l[0]),            
            'red'
        )
        assert l[0] != l[1]
        assert l[1] != l[0]

def enh_test():
    for row in ENHARMONIC_MATRIX:
        if len(row) == 3:
            l = [
                [ row[0], row[1] ],
                [ row[1], row[2] ],
                [ row[0], row[2] ],
            ]
        elif len(row) == 2:
            l = [
                [ row[0], row[1] ],
            ]
        equality_test(l)


def key_test(key_class):
    for _row in ENHARMONIC_MATRIX:
        for _enharmonic_note in _row:

            # ignore double alteration notes
            if len(_enharmonic_note.alt) >1:
                continue

            echo(
                '{} {}'.format(
                    key_class.__name__, _enharmonic_note
                ), 'blue'
            )
            echo(
                key_class(*_enharmonic_note),
                'cyan'
            )


equality_test(EQUALS)
inequality_test(NON_EQUALS)
enh_test()

### SCALES TESTS
key_test(ChromaticKey)
key_test(MajorKey)
key_test(NaturalMinorKey)
key_test(MelodicMinorKey)
key_test(HarmonicMinorKey)



# if __name__ == '__main__':

#     try:
#         ### NOTES TESTS
#         equality_test(EQUALS)
#         inequality_test(NON_EQUALS)
#         enh_test()

#         # ### SCALES TESTS
#         key_test(ChromaticKey)
#         key_test(MajorKey)
#         key_test(NaturalMinorKey)
#         key_test(MelodicMinorKey)
#         key_test(HarmonicMinorKey)

#     except KeyboardInterrupt:
#         pass

#     finally:
#         echo('Done', 'green')
